@@ -22,12 +22,14 @@ import (
 
 // Server holds the state of the Language Server.
 type Server struct {
-	server         *glspserv.Server
-	notebooks      *core.NotebookStore
-	documents      *documentStore
-	templateLoader core.TemplateLoader
-	fs             core.FileStorage
-	logger         util.Logger
+	server                 *glspserv.Server
+	notebooks              *core.NotebookStore
+	documents              *documentStore
+	noteContentParser      core.NoteContentParser
+	templateLoader         core.TemplateLoader
+	fs                     core.FileStorage
+	logger                 util.Logger
+	useAdditionalTextEdits opt.Bool
 }
 
 // ServerOpts holds the options to create a new Server.
@@ -59,12 +61,13 @@ func NewServer(opts ServerOpts) *Server {
 	}
 
 	server := &Server{
-		server:         glspServer,
-		notebooks:      opts.Notebooks,
-		documents:      newDocumentStore(fs, opts.Logger),
-		templateLoader: opts.TemplateLoader,
-		fs:             fs,
-		logger:         opts.Logger,
+		server:                 glspServer,
+		notebooks:              opts.Notebooks,
+		documents:              newDocumentStore(fs, opts.Logger),
+		templateLoader:         opts.TemplateLoader,
+		fs:                     fs,
+		logger:                 opts.Logger,
+		useAdditionalTextEdits: opt.NullBool,
 	}
 
 	var clientCapabilities protocol.ClientCapabilities
@@ -78,6 +81,15 @@ func NewServer(opts ServerOpts) *Server {
 			protocol.SetTraceValue(*params.Trace)
 		}
 
+		if params.ClientInfo != nil {
+			if params.ClientInfo.Name == "Visual Studio Code" {
+				// Visual Studio Code doesn't seem to support inl
+				// VSCode doesn't support deleting the trigger characters with
+				// the main TextEdit. We'll use additional text edits instead.
+				server.useAdditionalTextEdits = opt.True
+			}
+		}
+
 		capabilities := handler.CreateServerCapabilities()
 		capabilities.HoverProvider = true
 		capabilities.DefinitionProvider = true
@@ -178,8 +190,6 @@ func NewServer(opts ServerOpts) *Server {
 	}
 
 	handler.TextDocumentCompletion = func(context *glsp.Context, params *protocol.CompletionParams) (interface{}, error) {
-		// We don't use the context because clients might not send it. Instead,
-		// we'll look for trigger patterns in the document.
 		doc, ok := server.documents.Get(params.TextDocument.URI)
 		if !ok {
 			return nil, nil
@@ -190,28 +200,11 @@ func NewServer(opts ServerOpts) *Server {
 			return nil, err
 		}
 
-		switch doc.LookBehind(params.Position, 3) {
-		case "]((":
-			return server.buildLinkCompletionList(doc, notebook, params)
+		if params.Context != nil && params.Context.TriggerKind == protocol.CompletionTriggerKindInvoked {
+			return server.buildInvokedCompletionList(notebook, doc, params.Position)
+		} else {
+			return server.buildTriggerCompletionList(notebook, doc, params.Position)
 		}
-
-		switch doc.LookBehind(params.Position, 2) {
-		case "[[":
-			return server.buildLinkCompletionList(doc, notebook, params)
-		}
-
-		switch doc.LookBehind(params.Position, 1) {
-		case "#":
-			if notebook.Config.Format.Markdown.Hashtags {
-				return server.buildTagCompletionList(notebook, "#")
-			}
-		case ":":
-			if notebook.Config.Format.Markdown.ColonTags {
-				return server.buildTagCompletionList(notebook, ":")
-			}
-		}
-
-		return nil, nil
 	}
 
 	handler.CompletionItemResolve = func(context *glsp.Context, params *protocol.CompletionItem) (*protocol.CompletionItem, error) {
@@ -649,7 +642,45 @@ func (s *Server) refreshDiagnosticsOfDocument(doc *document, notify glsp.NotifyF
 	}()
 }
 
-func (s *Server) buildTagCompletionList(notebook *core.Notebook, triggerChar string) ([]protocol.CompletionItem, error) {
+// buildInvokedCompletionList builds the completion item response for a
+// completion started automatically when typing an identifier, or manually.
+func (s *Server) buildInvokedCompletionList(notebook *core.Notebook, doc *document, position protocol.Position) ([]protocol.CompletionItem, error) {
+	if !doc.IsTagPosition(position, notebook.Parser) {
+		return nil, nil
+	}
+	return s.buildTagCompletionList(notebook, doc.WordAt(position))
+}
+
+// buildTriggerCompletionList builds the completion item response for a
+// completion started with a trigger character.
+func (s *Server) buildTriggerCompletionList(notebook *core.Notebook, doc *document, position protocol.Position) ([]protocol.CompletionItem, error) {
+	// We don't use the context because clients might not send it. Instead,
+	// we'll look for trigger patterns in the document.
+	switch doc.LookBehind(position, 3) {
+	case "]((":
+		return s.buildLinkCompletionList(notebook, doc, position)
+	}
+
+	switch doc.LookBehind(position, 2) {
+	case "[[":
+		return s.buildLinkCompletionList(notebook, doc, position)
+	}
+
+	switch doc.LookBehind(position, 1) {
+	case "#":
+		if notebook.Config.Format.Markdown.Hashtags {
+			return s.buildTagCompletionList(notebook, "#")
+		}
+	case ":":
+		if notebook.Config.Format.Markdown.ColonTags {
+			return s.buildTagCompletionList(notebook, ":")
+		}
+	}
+
+	return nil, nil
+}
+
+func (s *Server) buildTagCompletionList(notebook *core.Notebook, prefix string) ([]protocol.CompletionItem, error) {
 	tags, err := notebook.FindCollections(core.CollectionKindTag, nil)
 	if err != nil {
 		return nil, err
@@ -659,7 +690,7 @@ func (s *Server) buildTagCompletionList(notebook *core.Notebook, triggerChar str
 	for _, tag := range tags {
 		items = append(items, protocol.CompletionItem{
 			Label:      tag.Name,
-			InsertText: s.buildInsertForTag(tag.Name, triggerChar, notebook.Config),
+			InsertText: s.buildInsertForTag(tag.Name, prefix, notebook.Config),
 			Detail:     stringPtr(fmt.Sprintf("%d %s", tag.NoteCount, strutil.Pluralize("note", tag.NoteCount))),
 		})
 	}
@@ -667,8 +698,8 @@ func (s *Server) buildTagCompletionList(notebook *core.Notebook, triggerChar str
 	return items, nil
 }
 
-func (s *Server) buildInsertForTag(name string, triggerChar string, config core.Config) *string {
-	switch triggerChar {
+func (s *Server) buildInsertForTag(name string, prefix string, config core.Config) *string {
+	switch prefix {
 	case ":":
 		name += ":"
 	case "#":
@@ -683,8 +714,8 @@ func (s *Server) buildInsertForTag(name string, triggerChar string, config core.
 	return &name
 }
 
-func (s *Server) buildLinkCompletionList(doc *document, notebook *core.Notebook, params *protocol.CompletionParams) ([]protocol.CompletionItem, error) {
-	linkFormatter, err := newLinkFormatter(doc, notebook, params)
+func (s *Server) buildLinkCompletionList(notebook *core.Notebook, doc *document, position protocol.Position) ([]protocol.CompletionItem, error) {
+	linkFormatter, err := newLinkFormatter(notebook, doc, position)
 	if err != nil {
 		return nil, err
 	}
@@ -701,7 +732,7 @@ func (s *Server) buildLinkCompletionList(doc *document, notebook *core.Notebook,
 
 	var items []protocol.CompletionItem
 	for _, note := range notes {
-		item, err := s.newCompletionItem(notebook, note, doc, params.Position, linkFormatter, templates)
+		item, err := s.newCompletionItem(notebook, note, doc, position, linkFormatter, templates)
 		if err != nil {
 			s.logger.Err(err)
 			continue
@@ -713,8 +744,8 @@ func (s *Server) buildLinkCompletionList(doc *document, notebook *core.Notebook,
 	return items, nil
 }
 
-func newLinkFormatter(doc *document, notebook *core.Notebook, params *protocol.CompletionParams) (core.LinkFormatter, error) {
-	if doc.LookBehind(params.Position, 3) == "]((" {
+func newLinkFormatter(notebook *core.Notebook, doc *document, position protocol.Position) (core.LinkFormatter, error) {
+	if doc.LookBehind(position, 3) == "]((" {
 		return core.NewMarkdownLinkFormatter(notebook.Config.Format.Markdown, true)
 	} else {
 		return notebook.NewLinkFormatter()
@@ -772,24 +803,30 @@ func (s *Server) newCompletionItem(notebook *core.Notebook, note core.MinimalNot
 		return item, err
 	}
 
-	addTextEdits := []protocol.TextEdit{}
+	if s.useAdditionalTextEditsWithNotebook(notebook) {
+		addTextEdits := []protocol.TextEdit{}
 
-	// Some LSP clients (e.g. VSCode) don't support deleting the trigger
-	// characters with the main TextEdit. So let's add an additional
-	// TextEdit for that.
-	addTextEdits = append(addTextEdits, protocol.TextEdit{
-		NewText: "",
-		Range:   rangeFromPosition(pos, -2, 0),
-	})
+		// Some LSP clients (e.g. VSCode) don't support deleting the trigger
+		// characters with the main TextEdit. So let's add an additional
+		// TextEdit for that.
+		addTextEdits = append(addTextEdits, protocol.TextEdit{
+			NewText: "",
+			Range:   rangeFromPosition(pos, -2, 0),
+		})
 
-	item.AdditionalTextEdits = addTextEdits
+		item.AdditionalTextEdits = addTextEdits
+	}
 
 	return item, nil
 }
 
 func (s *Server) newTextEditForLink(notebook *core.Notebook, note core.MinimalNote, doc *document, pos protocol.Position, linkFormatter core.LinkFormatter) (interface{}, error) {
-	currentDir := filepath.Dir(doc.Path)
-	context, err := core.NewLinkFormatterContext(note, notebook.Path, currentDir)
+	path := core.NotebookPath{
+		Path:       note.Path,
+		BasePath:   notebook.Path,
+		WorkingDir: filepath.Dir(doc.Path),
+	}
+	context, err := core.NewLinkFormatterContext(path, note.Title, note.Metadata)
 	if err != nil {
 		return nil, err
 	}
@@ -798,6 +835,12 @@ func (s *Server) newTextEditForLink(notebook *core.Notebook, note core.MinimalNo
 		return nil, err
 	}
 
+	// Overwrite [[ trigger directly if the additional text edits are disabled.
+	startOffset := 0
+	if !s.useAdditionalTextEditsWithNotebook(notebook) {
+		startOffset = -2
+	}
+
 	// Some LSP clients (e.g. VSCode) auto-pair brackets, so we need to
 	// remove the closing ]] or )) after the completion.
 	endOffset := 0
@@ -808,10 +851,17 @@ func (s *Server) newTextEditForLink(notebook *core.Notebook, note core.MinimalNo
 
 	return protocol.TextEdit{
 		NewText: link,
-		Range:   rangeFromPosition(pos, 0, endOffset),
+		Range:   rangeFromPosition(pos, startOffset, endOffset),
 	}, nil
 }
 
+func (s *Server) useAdditionalTextEditsWithNotebook(nb *core.Notebook) bool {
+	return nb.Config.LSP.Completion.UseAdditionalTextEdits.
+		Or(s.useAdditionalTextEdits).
+		OrBool(false).
+		Unwrap()
+}
+
 func positionInRange(content string, rng protocol.Range, pos protocol.Position) bool {
 	start, end := rng.IndexesIn(content)
 	i := pos.IndexIn(content)
